##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
try:
    from qiskit.providers import ProviderV1 as Provider
    from qiskit.providers.exceptions import QiskitBackendNotFoundError
except ImportError:
    raise ImportError(
    "Missing optional 'qiskit' dependencies. \
To install run: pip install azure-quantum[qiskit]"
    )

from typing import Dict, Iterable
from azure.quantum import Workspace

from azure.quantum.qiskit.job import AzureQuantumJob
from azure.quantum.qiskit.backends import *

QISKIT_USER_AGENT = "azure-quantum-qiskit"


class AzureQuantumProvider(Provider):
    def __init__(self, workspace=None, **kwargs):
        self._backends = None
        if workspace is None:
            workspace = Workspace(**kwargs)

        workspace.append_user_agent(QISKIT_USER_AGENT)

        self._workspace = workspace

    def get_workspace(self) -> Workspace:
        return self._workspace

    def _get_all_subclasses(self, cls):
        all_subclasses = []
        for subclass in cls.__subclasses__():
            all_subclasses.append(subclass)
            all_subclasses.extend(self._get_all_subclasses(subclass))
        return all_subclasses

    def get_backend(self, name=None, **kwargs):
        """
        Return a single backend matching the specified filtering.
        """
        backends = self.backends(name, **kwargs)
        if len(backends) > 1:
            raise QiskitBackendNotFoundError("More than one backend matches the criteria")
        if not backends:
            raise QiskitBackendNotFoundError(f"Could not find target '{name}'. \
Please make sure the target name is valid and that the associated provider is added to your Workspace. \
To add a provider to your quantum workspace on the Azure Portal, \
see https://aka.ms/AQ/Docs/AddProvider")
        return backends[0]

    def backends(self, name=None, provider_id=None, **kwargs):
        """Return a list of backends matching the specified filtering.
        Args:
            name (str): name of the backend.
            **kwargs: dict used for filtering.
        Returns:
            list[Backend]: a list of Backends that match the filtering
                criteria.
        """
        from azure.quantum.target.target_factory import TargetFactory
        from qiskit.providers import BackendV1 as Backend
        from azure.quantum.qiskit.backends import DEFAULT_TARGETS

        all_targets = {
            name: _t for t in self._get_all_subclasses(Backend)
            for _t in [t]
            if hasattr(_t, "backend_names")
            for name in _t.backend_names
        }

        target_factory = TargetFactory(
            base_cls=Backend,
            workspace=self._workspace,
            default_targets=DEFAULT_TARGETS,
            all_targets=all_targets
        )

        targets = target_factory.get_targets(
            name=name,
            provider_id=provider_id,
            provider=self,
            **kwargs
        )

        # Always return an iterable
        if isinstance(targets, Iterable):
            return targets
        return [targets]

    def get_job(self, job_id) -> AzureQuantumJob:
        """ Returns the Job instance associated with the given id."""
        azure_job = self._workspace.get_job(job_id)
        backend = self.get_backend(azure_job.details.target)
        return AzureQuantumJob(backend, azure_job)



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
from typing import TYPE_CHECKING, Dict, Sequence

if TYPE_CHECKING:
    import cirq
    from azure.quantum import Job as AzureJob


class Job:
    """
    Thin wrapper around an Azure Quantum Job that supports
    returning results in Cirq format.
    """
    def __init__(
        self,
        azure_job: "AzureJob",
        program: "cirq.Circuit",
        measurement_dict: dict = None
    ):
        """Construct a Job."""
        self._azure_job = azure_job
        self._program = program
        self._measurement_dict = measurement_dict

    def job_id(self) -> str:
        """Returns the job id (UID) for the job."""
        return self._azure_job.id

    def status(self) -> str:
        """Gets the current status of the job."""
        self._azure_job.refresh()
        status = self._azure_job.details.status
        if status == "Failed":
            return f"{status}: {self._azure_job.details.error_data.message}"
        else:
            return status

    def target(self) -> str:
        """Returns the target where the job was run."""
        return self._azure_job.details.target

    def name(self) -> str:
        """Returns the name of the job which was supplied during job creation."""
        return self._azure_job.details.name

    def num_qubits(self) -> int:
        """Returns the number of qubits for the job."""
        return self._azure_job.details.metadata["qubits"]

    def repetitions(self) -> int:
        """Returns the number of repetitions for the job."""
        return self._azure_job.details.metadata["repetitions"]

    def measurement_dict(self) -> Dict[str, Sequence[int]]:
        """Returns a dictionary of measurement keys to target qubit index."""
        if self._measurement_dict is None:
            from cirq import MeasurementGate
            measurements = [op for op in self._program.all_operations() if isinstance(op.gate, MeasurementGate)]
            self._measurement_dict = {
                meas.gate.key: [q.x for q in meas.qubits] for meas in measurements
            }
        return self._measurement_dict

    def results(self, timeout_seconds: int = 7200) -> "cirq.Result":
        """Poll the Azure Quantum API for results."""
        return self._azure_job.get_results(timeout_secs=timeout_seconds)

    def cancel(self):
        """Cancel the given job."""
        self._azure_job.workspace.cancel_job(self._azure_job)

    def delete(self):
        """Delete the given job."""
        self._azure_job.workspace.cancel_job(self._azure_job)

    def __str__(self) -> str:
        return f'azure.quantum.cirq.Job(job_id={self.job_id()})'



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import logging
from multiprocessing.sharedctypes import Value

from typing import Optional

from azure.quantum.target.solvers import Solver
from azure.quantum.target.solvers import HardwarePlatform
from azure.quantum.workspace import Workspace
from azure.quantum._client.models import TargetStatus

logger = logging.getLogger(__name__)


class SimulatedAnnealing(Solver):
    target_names = [
        "microsoft.simulatedannealing-parameterfree.fpga",
        "microsoft.simulatedannealing.fpga",
        "microsoft.simulatedannealing.cpu",
        "microsoft.simulatedannealing-parameterfree.cpu",
        "microsoft.simulatedannealing.cpu.experimental",
        "microsoft.simulatedannealing-parameterfree.cpu.experimental"
    ]
    def __init__(
        self,
        workspace: Workspace,
        name: str = "microsoft.simulatedannealing.cpu",
        beta_start: Optional[float] = None,
        beta_stop: Optional[float] = None,
        sweeps: Optional[int] = None,
        restarts: Optional[int] = None,
        timeout: Optional[int] = None,
        seed: Optional[int] = None,
        platform: Optional[HardwarePlatform] = HardwarePlatform.CPU,
        **kwargs
    ):
        """The constructor of an Simulated Annealing solver.

        Multi-core Simulated Annealing solver for binary optimization problems
        with k-local interactions on an all-to-all graph topology with double
        precision support for the coupler weights.

        :param beta_start:
            specifies the starting inverse temperature.
        :param beta_stop:
            specifies the stopping inverse temperature.
        :param sweeps:
            specifies the number of sweeps.
        :param restarts:
            specifies the number of restarts.
        :param timeout:
            specifies maximum number of seconds to run the core solver
            loop. initialization time does not respect this value, so the
            solver may run longer than the value specified.
        :param seed:
            specifies a random seed value.
        :platform:
            specifies hardware platform
            HardwarePlatform.CPU or HardwarePlatform.FPGA.
        """
        param_free = (
            beta_start is None
            and beta_stop is None
            and sweeps is None
            and restarts is None
        )

        if platform == HardwarePlatform.FPGA:
            name = (
                "microsoft.simulatedannealing-parameterfree.fpga"
                if param_free
                else "microsoft.simulatedannealing.fpga"
            )
        elif param_free:
            if "experimental" in name:
                name = "microsoft.simulatedannealing-parameterfree.cpu.experimental"
            else:
                name = "microsoft.simulatedannealing-parameterfree.cpu"

        super().__init__(
            workspace=workspace,
            provider_id="Microsoft",
            name=name,
            input_data_format="microsoft.qio.v2",
            output_data_format="microsoft.qio-results.v2",
            **kwargs
        )

        self.set_one_param("beta_start", beta_start)
        self.set_one_param("beta_stop", beta_stop)
        self.set_one_param("sweeps", sweeps)
        self.set_one_param("restarts", restarts)
        self.set_one_param("timeout", timeout)
        self.set_one_param("seed", seed)

    @classmethod
    def from_target_status(
        cls, workspace: "Workspace", status: TargetStatus, **kwargs
    ):
        """Create a SimulatedAnnealing instance from a given workspace and target status.

        :param workspace: Associated workspace
        :type workspace: Workspace
        :param status: Target status with availability and current queue time
        :type status: TargetStatus
        :return: Target instance
        :rtype: Target
        """
        if status.id.endswith("cpu") or status.id.endswith("cpu.experimental"):
            platform = HardwarePlatform.CPU
        elif status.id.endswith("fpga"):
            platform = HardwarePlatform.FPGA
        else:
            raise ValueError(f"SimulatedAnnealing solver with name {status.id} is not supported.")

        return super().from_target_status(
            workspace=workspace,
            status=status,
            platform=platform,
            **kwargs
        )

    def supports_grouped_terms(self):
        if "experimental" in self.name:
            return True
        return False
    
    def supports_protobuf(self):
        if "experimental" in self.name:
            return True
        return False



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
from azure.quantum.aio.target.solvers import Solver
from azure.quantum.target.microsoft.qio import ParallelTempering as SyncParallelTempering
from azure.quantum.target.microsoft.qio import SimulatedAnnealing as SyncSimulatedAnnealing
from azure.quantum.target.microsoft.qio import Tabu as SyncTabu
from azure.quantum.target.microsoft.qio import QuantumMonteCarlo as SyncQuantumMonteCarlo
from azure.quantum.target.microsoft.qio import PopulationAnnealing as SyncPopulationAnnealing
from azure.quantum.target.microsoft.qio import SubstochasticMonteCarlo as SyncSubstochasticMonteCarlo


class ParallelTempering(Solver, SyncParallelTempering):
    pass


class SimulatedAnnealing(Solver, SyncSimulatedAnnealing):
    pass


class Tabu(Solver, SyncTabu):
    pass


class QuantumMonteCarlo(Solver, SyncQuantumMonteCarlo):
    pass


class PopulationAnnealing(Solver, SyncPopulationAnnealing):
    pass


class SubstochasticMonteCarlo(Solver, SyncSubstochasticMonteCarlo):
    pass



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from enum import Enum
from six import with_metaclass
from azure.core import CaseInsensitiveEnumMeta


class DimensionScope(with_metaclass(CaseInsensitiveEnumMeta, str, Enum)):
    """The scope at which the quota is applied.
    """

    WORKSPACE = "Workspace"
    SUBSCRIPTION = "Subscription"

class JobStatus(with_metaclass(CaseInsensitiveEnumMeta, str, Enum)):
    """The job status.
    """

    WAITING = "Waiting"
    EXECUTING = "Executing"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    CANCELLED = "Cancelled"

class JsonPatchOperation(with_metaclass(CaseInsensitiveEnumMeta, str, Enum)):
    """The operation to be performed.
    """

    ADD = "add"
    REMOVE = "remove"
    REPLACE = "replace"
    MOVE = "move"
    COPY = "copy"
    TEST = "test"

class MeterPeriod(with_metaclass(CaseInsensitiveEnumMeta, str, Enum)):
    """The time period in which the quota's underlying meter is accumulated. Based on calendar year.
    'None' is used for concurrent quotas.
    """

    NONE = "None"
    MONTHLY = "Monthly"

class ProviderAvailability(with_metaclass(CaseInsensitiveEnumMeta, str, Enum)):
    """Provider availability.
    """

    AVAILABLE = "Available"
    DEGRADED = "Degraded"
    UNAVAILABLE = "Unavailable"

class TargetAvailability(with_metaclass(CaseInsensitiveEnumMeta, str, Enum)):
    """Target availability.
    """

    AVAILABLE = "Available"
    DEGRADED = "Degraded"
    UNAVAILABLE = "Unavailable"



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import warnings
import asyncio
from typing import Any, Dict, List, TYPE_CHECKING, Union
from azure.quantum.target import *

if TYPE_CHECKING:
    from azure.quantum import Workspace
    from azure.quantum._client.models import TargetStatus

# Target ID keyword for parameter-free solvers
PARAMETER_FREE = "parameterfree"


class TargetFactory:
    """Factory class for generating a Target based on a
    provider and target name
    """
    __instances = {}

    def __new__(cls, *args, **kwargs):
        base_cls = kwargs.get("base_cls")
        if cls.__instances.get(base_cls) is None:
            cls.__instances[base_cls] = super().__new__(cls)
        return cls.__instances[base_cls]

    def __init__(
        self,
        base_cls: Target,
        workspace: "Workspace",
        default_targets: Dict[str, Any] = DEFAULT_TARGETS,
        all_targets: Dict[str, Any] = None
    ):
        """Target factory class for creating targets
        based on a name and/or provider ID.

        :param base_cls: Base class for findng first and second 
            generation child classes.
        :type base_cls: Target
        :param workspace: Azure Quantum Workspace
        :type workspace: Workspace
        :param default_targets: Dictionary of default target classes keyed
            by provider ID, defaults to DEFAULT_TARGETS
        :type default_targets: Dict[str, Any], optional
        :param all_targets: Dictionary of all target classes by name,
            optional. Defaults to finding all first and second degree 
            subclasses of base_cls by name via cls.target_names.
        :type all_targets: Dict[str, Any]
        """
        self._workspace = workspace
        self._base_cls = base_cls
        # case insensitive lookup
        self._default_targets = {k.lower(): v for k, v in default_targets.items()}
        self._all_targets = all_targets or self._get_all_target_cls()

    def _get_all_target_cls(self) -> Dict[str, Target]:
        """Get all target classes by target name"""
        return {
            name.lower(): _t for t in self._base_cls.__subclasses__()
            for _t in t.__subclasses__() + [t]
            if hasattr(_t, "target_names")
            for name in _t.target_names
            if not asyncio.iscoroutinefunction(_t.submit)
        }

    def _target_cls(self, provider_id: str, name: str):
        if name in self._all_targets:
            return self._all_targets[name.lower()]

        if provider_id.lower() in self._default_targets:
            return self._default_targets[provider_id.lower()]
 
        warnings.warn(
            f"No default target specified for provider {provider_id}. \
Please check the provider name and try again or create an issue here: \
https://github.com/microsoft/qdk-python/issues.")
        return Target

    def create_target(
        self, provider_id: str, name: str, **kwargs
    ) -> Target:
        """Create target from provider ID and target name.

        :param workspace: Workspace
        :type workspace: Workspace
        :param provider_id: Provider name
        :type provider_id: str
        :param name: Target name
        :type name: str
        :return: Target instance
        :rtype: Target
        """
        cls = self._target_cls(provider_id, name)
        if cls is not None:
            return cls(
                workspace=self._workspace,
                name=name,
                provider_id=provider_id,
                **kwargs
            )

    def from_target_status(
        self,
        provider_id: str,
        status: "TargetStatus",
        **kwargs
    ):
        cls = self._target_cls(provider_id, status.id)
        if hasattr(cls, "from_target_status"):
            return cls.from_target_status(self._workspace, status, **kwargs)
        elif cls is not None:
            return cls(name=status.id, **kwargs)

    def get_targets(
        self,
        name: str,
        provider_id: str,
        **kwargs
    ) -> Union[Target, List[Target]]:
        """Create targets that are available to this workspace
        filtered by name and provider ID.

        :param name: Target name
        :type name: str
        :param workspace: Workspace
        :type workspace: Workspace
        :param provider_id: Provider name
        :type provider_id: str
        :return: One or more Target objects
        :rtype: Union[Target, List[Target]]
        """
        target_statuses = self._workspace._get_target_status(name, provider_id)

        if len(target_statuses) == 1:
            return self.from_target_status(*target_statuses[0], **kwargs)

        else:
            # Don't return redundant parameter-free targets
            return [
                self.from_target_status(_provider_id, status, **kwargs)
                for _provider_id, status in target_statuses
                if PARAMETER_FREE not in status.id
                and (
                    _provider_id.lower() in self._default_targets
                    or status.id in self._all_targets
                )
            ]



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
import warnings
import json

import logging
logger = logging.getLogger(__name__)

from typing import TYPE_CHECKING, Union, List
from azure.quantum.version import __version__
from azure.quantum.qiskit.job import AzureQuantumJob

try:
    from qiskit import QuantumCircuit
    from qiskit.providers import BackendV1 as Backend
    from qiskit.qobj import Qobj, QasmQobj

except ImportError:
    raise ImportError(
    "Missing optional 'qiskit' dependencies. \
To install run: pip install azure-quantum[qiskit]"
)

# Set of gates supported by QIR targets.
QIR_BASIS_GATES = [
    "x",
    "y",
    "z",
    "rx",
    "ry",
    "rz",
    "h",
    "cx",
    "cz",
    "s",
    "sdg",
    "t",
    "tdg",
    "measure",
    "reset"
]

class AzureBackend(Backend):
    """Base class for interfacing with an IonQ backend in Azure Quantum"""
    backend_name = None

    def _prepare_job_metadata(self, circuit):
        """ Returns the metadata relative to the given circuit that will be attached to the Job"""
        return {
            "qiskit": True,
            "name": circuit.name,
            "num_qubits": circuit.num_qubits,
            "metadata": json.dumps(circuit.metadata),
        }

    def _translate_input(self, circuit, data_format, input_params):
        """ Translates the input values to the format expected by the AzureBackend. """
        if data_format != "qir.v1":
            target = self.name()
            raise ValueError(f"{data_format} is not a supported data format for target {target}.")

        logger.info(f"Using QIR as the job's payload format.")
        from qiskit_qir import to_qir_bitcode, to_qir

        capability = input_params["targetCapability"] if "targetCapability" in input_params else "AdaptiveProfileExecution"

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"QIR:\n{to_qir(circuit, capability)}")

        # all qir payload needs to define an entryPoint and arguments:
        if not "entryPoint" in input_params:
            input_params["entryPoint"] = "main"
        if not "arguments" in input_params:
            input_params["arguments"] = []

        qir = bytes(to_qir_bitcode(circuit, capability))
        return (qir, data_format, input_params)

    def run(self, circuit, **kwargs):
        """Submits the given circuit to run on an Azure Quantum backend."""        

        # Some Qiskit features require passing lists of circuits, so unpack those here.
        # We currently only support single-experiment jobs.
        if isinstance(circuit, (list, tuple)):
            if len(circuit) > 1:
                raise NotImplementedError("Multi-experiment jobs are not supported!")
            circuit = circuit[0]

        # If the circuit was created using qiskit.assemble,
        # disassemble into QASM here
        if isinstance(circuit, QasmQobj) or isinstance(circuit, Qobj):
            from qiskit.assembler import disassemble
            circuits, run, _ = disassemble(circuit)
            circuit = circuits[0]
            if kwargs.get("shots") is None:
                # Note that the default number of shots for QObj is 1024
                # unless the user specifies the backend.
                kwargs["shots"] = run["shots"]
        
        # The default of these job parameters come from the AzureBackend configuration:
        config = self.configuration()
        blob_name = kwargs.pop("blob_name", config.azure["blob_name"])
        content_type = kwargs.pop("content_type", config.azure["content_type"])
        provider_id = kwargs.pop("provider_id", config.azure["provider_id"])
        input_data_format = kwargs.pop("input_data_format", config.azure["input_data_format"])
        output_data_format = kwargs.pop("output_data_format", config.azure["output_data_format"])

        # If not provided as kwargs, the values of these parameters 
        # are calculated from the circuit itself:
        job_name = kwargs.pop("job_name", circuit.name)
        metadata = kwargs.pop("metadata") if "metadata" in kwargs else self._prepare_job_metadata(circuit)

        # Backend options are mapped to input_params.
        # Take also into consideration options passed in the kwargs, as the take precedence
        # over default values:
        input_params = vars(self.options)
        for opt in kwargs.copy():
            if opt in input_params:
                input_params[opt] = kwargs.pop(opt)

        # Some providers refer as 'shots' the 'count' parameter,
        # Remove this once all providers accept "count":
        if "shots" in input_params:
            input_params["count"] = input_params["shots"]

        # translate
        (input_data, input_data_format, input_params) = self._translate_input(circuit, input_data_format, input_params)

        logger.info(f"Submitting new job for backend {self.name()}")
        job = AzureQuantumJob(
            backend=self,
            target=self.name(),
            name=job_name,
            input_data=input_data,
            blob_name=blob_name,
            content_type=content_type,
            provider_id=provider_id,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            input_params = input_params,
            metadata=metadata,
            **kwargs
        )

        logger.info(f"Submitted job with id '{job.id()}' for circuit '{circuit.name}':")
        logger.info(input_data)

        return job

    def retrieve_job(self, job_id) -> AzureQuantumJob:
        """ Returns the Job instance associated with the given id."""
        return self._provider.get_job(job_id)



from azure.quantum.target.oneqbit import (
    TabuSearch as SyncTabuSearch,
    PticmSolver as SyncPticmSolver,
    PathRelinkingSolver as SyncPathRelinkingSolver,
)
from azure.quantum.aio.target.solvers import Solver


class TabuSearch(SyncTabuSearch, Solver):
    pass


class PticmSolver(SyncPticmSolver, Solver):
    pass


class PathRelinkingSolver(SyncPathRelinkingSolver, Solver):
    pass



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
import warnings

from typing import TYPE_CHECKING, Union, List
from azure.quantum.version import __version__
from azure.quantum.qiskit.job import AzureQuantumJob

from .backend import AzureBackend

from qiskit import QuantumCircuit
from qiskit.providers.models import BackendConfiguration
from qiskit.providers import Options

if TYPE_CHECKING:
    from azure.quantum.qiskit import AzureQuantumProvider

import logging
logger = logging.getLogger(__name__)

__all__ = [
    "QuantinuumBackend",
    "QuantinuumQPUBackend",
    "QuantinuumAPIValidatorBackend",
    "QuantinuumSimulatorBackend"
]

QUANTINUUM_BASIS_GATES = [
    "x",
    "y",
    "z",
    "rx",
    "ry",
    "rz",
    "h",
    "cx",
    "ccx",
    "cz",
    "s",
    "sdg",
    "t",
    "tdg",
    "v",
    "vdg",
    "zz",
    "measure",
    "reset"
]

QUANTINUUM_PROVIDER_ID = "quantinuum"
QUANTINUUM_PROVIDER_NAME = "Quantinuum"

HONEYWELL_PROVIDER_ID = "honeywell"
HONEYWELL_PROVIDER_NAME = "Honeywell"

class QuantinuumBackend(AzureBackend):
    """Base class for interfacing with a Quantinuum (formerly Quantinuum) backend in Azure Quantum"""

    def __init__(self, **kwargs):
        self._provider_id = QUANTINUUM_PROVIDER_ID
        self._provider_name = QUANTINUUM_PROVIDER_NAME
        if kwargs.pop("provider_id", None) == HONEYWELL_PROVIDER_ID:
            self._provider_id = HONEYWELL_PROVIDER_ID
            self._provider_name = HONEYWELL_PROVIDER_NAME
        super().__init__(**kwargs)

    @classmethod
    def _default_options(cls):
        return Options(count=500)

    def _azure_config(self):
        return {
            "blob_name": "inputData",
            "content_type": "application/qasm",
            "provider_id": self._provider_id,
            "input_data_format": "honeywell.openqasm.v1",
            "output_data_format": "honeywell.quantum-results.v1",
        }

    def _translate_input(self, circuit, data_format, input_params):
        """ Translates the input values to the format expected by the AzureBackend. """
        if data_format == "honeywell.openqasm.v1":
            return (circuit.qasm(), data_format, input_params)
        else:
            super()._translate_input(circuit, data_format, input_params)
        

    def estimate_cost(self, circuit: QuantumCircuit, shots: int = None, count: int = None):
        """Estimate cost for running this circuit

        :param circuit: Qiskit quantum circuit
        :type circuit: QuantumCircuit
        :param shots: Shot count
        :type shots: int
        :param count: Shot count (deprecated)
        :type count: int
        """
        if count is not None:
            shots = count
            warnings.warn(
                "Input parameter 'count' has been deprecated. Please use 'shots' instead.")
        if shots is None:
            raise ValueError("Missing input argument 'shots'.")

        input_data = circuit.qasm()
        workspace = self.provider().get_workspace()
        target = workspace.get_targets(self.name())
        return target.estimate_cost(input_data, num_shots=shots)


class QuantinuumAPIValidatorBackend(QuantinuumBackend):
    backend_names = (
        "quantinuum.hqs-lt-s1-apival",
        "quantinuum.hqs-lt-s2-apival"
    )

    def __init__(
        self,
        name: str,
        provider: "AzureQuantumProvider",
        **kwargs
    ):
        self._provider_id = QUANTINUUM_PROVIDER_ID
        self._provider_name = QUANTINUUM_PROVIDER_NAME
        if kwargs.pop("provider_id", None) == "honeywell":
            self._provider_id = HONEYWELL_PROVIDER_ID
            self._provider_name = HONEYWELL_PROVIDER_NAME

        default_config = BackendConfiguration.from_dict(
            {
                "backend_name": name,
                "backend_version": __version__,
                "simulator": True,
                "local": False,
                "coupling_map": None,
                "description": f"Quantinuum (formerly Honeywell) API validator on Azure Quantum",
                "basis_gates": QUANTINUUM_BASIS_GATES,
                "memory": False,
                "n_qubits": 10,
                "conditional": False,
                "max_shots": None,
                "max_experiments": 1,
                "open_pulse": False,
                "gates": [{"name": "TODO", "parameters": [], "qasm_def": "TODO"}],
                "azure": self._azure_config(),
            }
        )
        configuration: BackendConfiguration = kwargs.pop("configuration", default_config)
        logger.info(f"Initializing {self._provider_name}APIValidatorBackend")
        super().__init__(configuration=configuration,
                         provider=provider,
                         provider_id=self._provider_id,
                         **kwargs)


class QuantinuumSimulatorBackend(QuantinuumBackend):
    backend_names = (
        "quantinuum.hqs-lt-s1-sim",
        "quantinuum.hqs-lt-s2-sim"
    )

    def __init__(
        self,
        name: str,
        provider: "AzureQuantumProvider",
        **kwargs
    ):
        self._provider_id = QUANTINUUM_PROVIDER_ID
        self._provider_name = QUANTINUUM_PROVIDER_NAME
        if kwargs.pop("provider_id", None) == "honeywell":
            self._provider_id = HONEYWELL_PROVIDER_ID
            self._provider_name = HONEYWELL_PROVIDER_NAME

        configuration: BackendConfiguration = kwargs.pop("configuration", None)
        default_config = BackendConfiguration.from_dict(
            {
                "backend_name": name,
                "backend_version": __version__,
                "simulator": True,
                "local": False,
                "coupling_map": None,
                "description": f"Quantinuum (formerly Honeywell) simulator on Azure Quantum",
                "basis_gates": QUANTINUUM_BASIS_GATES,
                "memory": False,
                "n_qubits": 10,
                "conditional": False,
                "max_shots": None,
                "max_experiments": 1,
                "open_pulse": False,
                "gates": [{"name": "TODO", "parameters": [], "qasm_def": "TODO"}],
                "azure": self._azure_config(),
            }
        )
        configuration: BackendConfiguration = kwargs.pop("configuration", default_config)
        logger.info(f"Initializing {self._provider_name}APIValidatorBackend")
        super().__init__(configuration=configuration,
                         provider=provider,
                         provider_id=self._provider_id,
                         **kwargs)


class QuantinuumQPUBackend(QuantinuumBackend):
    backend_names = (
        "quantinuum.hqs-lt-s1",
        "quantinuum.hqs-lt-s2"
    )

    def __init__(
        self,
        name: str,
        provider: "AzureQuantumProvider",
        **kwargs
    ):
        self._provider_id = QUANTINUUM_PROVIDER_ID
        self._provider_name = QUANTINUUM_PROVIDER_NAME
        if kwargs.pop("provider_id", None) == "honeywell":
            self._provider_id = HONEYWELL_PROVIDER_ID
            self._provider_name = HONEYWELL_PROVIDER_NAME

        default_config = BackendConfiguration.from_dict(
            {
                "backend_name": name,
                "backend_version": __version__,
                "simulator": False,
                "local": False,
                "coupling_map": None,
                "description": f"Quantinuum (formerly Honeywell) QPU on Azure Quantum",
                "basis_gates": QUANTINUUM_BASIS_GATES,
                "memory": False,
                "n_qubits": 10,
                "conditional": False,
                "max_shots": 10000,
                "max_experiments": 1,
                "open_pulse": False,
                "gates": [{"name": "TODO", "parameters": [], "qasm_def": "TODO"}],
                "azure": self._azure_config(),
            }
        )
        configuration: BackendConfiguration = kwargs.pop("configuration", default_config)
        logger.info(f"Initializing {self._provider_name}QPUBackend")
        super().__init__(configuration=configuration,
                         provider=provider, 
                         provider_id=self._provider_id,
                         **kwargs)



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import logging

from typing import Optional

from azure.quantum.target.solvers import Solver
from azure.quantum.workspace import Workspace

logger = logging.getLogger(__name__)

class QuantumMonteCarlo(Solver):
    target_names = (
        "microsoft.qmc.cpu",
        "microsoft.qmc.cpu.experimental"
    )
    def __init__(
        self,
        workspace: Workspace,
        name: str = "microsoft.qmc.cpu",
        trotter_number: Optional[int] = None,
        seed: Optional[int] = None,
        transverse_field_start: Optional[float] = None,
        transverse_field_stop: Optional[float] = None,
        restarts: Optional[int] = None,
        sweeps: Optional[int] = None,
        beta_start: Optional[float] = None,
        **kwargs
    ):
        """The constructor of Simulated Qunatum Annelaing Search solver.

        Simulated Quantum Search solver for binary optimization problems
        with k-local interactions on an all-to-all graph topology with double
        precision support for the coupler weights.

        This solver is CPU only.

        :param trotter_number:
            specifies the number of trotter time slices
            i.e. number of copies of each variable.
        :param seed:
            specifies a random seed value.
        :param sweeps:
            Number of monte carlo sweeps
        :param transverse_field_start:
            starting strength of the external field
        :param transverse_field_end:
            ending strength of the external field
        :param beta_start:
            Low starting temp i.e. beta start
        :param restarts:
            Number of simulation runs
        """

        super().__init__(
            workspace=workspace,
            provider_id="Microsoft",
            name=name,
            input_data_format="microsoft.qio.v2",
            output_data_format="microsoft.qio-results.v2",
            **kwargs
        )
        self.set_one_param("sweeps", sweeps)
        self.set_one_param("trotter_number", trotter_number)
        self.set_one_param("seed", seed)
        self.set_one_param("transverse_field_start", transverse_field_start)
        self.set_one_param("transverse_field_stop", transverse_field_stop)
        self.set_one_param("beta_start", beta_start)
        self.set_one_param("restarts", restarts)

    def supports_grouped_terms(self):
        if "experimental" in self.name:
            return True
        return False
    
    def supports_protobuf(self):
        if "experimental" in self.name:
            return True
        return False



# coding=utf-8
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import warnings
warnings.warn("The azure.quantum.optimization.solvers namespace will be deprecated. \
Please use azure.quantum.target instead.")

from azure.quantum.target.solvers import Solver
from azure.quantum.target.microsoft.qio import (
    ParallelTempering,
    SimulatedAnnealing,
    Tabu,
    QuantumMonteCarlo,
    PopulationAnnealing,
    SubstochasticMonteCarlo,
)



#!/bin/env python
# -*- coding: utf-8 -*-
##
# setup.py: Installs Python host functionality for azure-quantum.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

# IMPORTS #

import setuptools
import os
import distutils

# VERSION INFORMATION #
# Our build process sets the PYTHON_VERSION environment variable to a version
# string that is compatible with PEP 440, and so we inherit that version number
# here and propagate that to qsharp/version.py.
#
# To make sure that local builds still work without the same environment
# variables, we'll default to 0.0.0.1 as a development version.

version = os.environ.get("PYTHON_VERSION", "0.0.0.1")

with open("./azure/quantum/version.py", "w") as f:
    f.write(
        f"""# Auto-generated file, do not edit.
##
# version.py: Specifies the version of the azure.quantum package.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
__version__ = "{version}"
"""
    )

with open("./azure/quantum/_client/_version.py", "w") as f:
    f.write(
        f"""# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

VERSION = "{version}"
"""
    )

# DESCRIPTION #
# The long description metadata passed to setuptools is used to populate the
# PyPI page for this package. Thus, we'll generate the description by using the
# same README.md file that we use in the GitHub repo.

with open("./README.md", "r") as fh:
    long_description = fh.read()

# LIST OF REQUIREMENTS #
# Get list of requirements from requirements.txt
with open("./requirements.txt", "r") as fh:
    requirements = fh.readlines()

# SETUPTOOLS INVOCATION #
setuptools.setup(
    name="azure-quantum",
    version=version,
    author="Microsoft",
    description="Python client for Azure Quantum",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/microsoft/qdk-python",
    packages=setuptools.find_namespace_packages(include=["azure.*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    extras_require={
        "qiskit": [
            "qiskit-ionq>=0.1.4",
            "qiskit-terra>=0.19.1",
            "qiskit-qir>=0.1.0b8"
        ],
        "cirq": [
            "cirq-core==0.14.1",
            "cirq-ionq==0.14.1",
        ]
    }
)



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
from .target import Target
from .solvers import Solver
from .microsoft.qio import (
    ParallelTempering,
    SimulatedAnnealing,
    Tabu,
    QuantumMonteCarlo,
    PopulationAnnealing,
    SubstochasticMonteCarlo,
)
from .oneqbit import (
    TabuSearch,
    PticmSolver,
    PathRelinkingSolver,
)
from .toshiba import (
    SimulatedBifurcationMachine
)
from .ionq import IonQ
from .honeywell import Honeywell
from .quantinuum import Quantinuum
from .rigetti import Rigetti

# Default targets to use when there is no target class
# associated with a given target ID
DEFAULT_TARGETS = {
    "ionq": IonQ,
    "honeywell": Honeywell,
    "quantinuum": Quantinuum,
    "rigetti": Rigetti,
    "Microsoft": Solver,
    "toshiba": Solver,
    "1qbit": Solver,
}



"""Defines targets and helper functions for the Rigetti provider"""

##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

__all__ = [
    "InputParams",
    "Readout",
    "Result",
    "Rigetti",
    "RigettiTarget",
]

from .result import Readout, Result
from .target import InputParams, Rigetti, RigettiTarget



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import logging
import warnings

from typing import Optional, List
from azure.quantum import Workspace
from azure.quantum.target import Solver

logger = logging.getLogger(__name__)

__all__ = ["TabuSearch", "PticmSolver", "PathRelinkingSolver"]


class TabuSearch(Solver):
    target_names = ("1qbit.tabu",)

    def __init__(
        self,
        workspace: Workspace,
        name: str = "1qbit.tabu",
        improvement_cutoff: Optional[int] = None,
        improvement_tolerance: Optional[float] = None,
        seed: Optional[int] = None,
        tabu_tenure: Optional[int] = None,
        tabu_tenure_rand_max: Optional[int] = None,
        timeout: Optional[int] = None,
        **kwargs
    ):
        """The constructor of a Tabu Search solver.

        An iterative heuristic algorithm that uses local search techniques
        to solve a QUBO problem. It starts from a random solution and looks
        for an improved solution in the solution's neighborhood which includes
        all possible single flips.
        The algorithm stops when it reaches a stopping
        criterion, such as a specified number of consecutive iterations without
        improvement.

        For more information please visit:
        https://portal.1qbit.com/docs/model/tabusolver

        :param improvement_cutoff:
            The number of iterations that
            the solver attempts with no improvement
            before stopping. Default: 0
        :param improvement_tolerance:
            The tolerance value that determines if a solution is an improvement
            over the previous iteration. Default: 1e-9
        :param seed:
            The seed in the random number generator to control for deterministic
            results. Default: a random integer
        :param tabu_tenure:
            The tenure prevents a flipped
            variable from being flipped again during
            the iterations. Default: 0
        :param tabu_tenure_rand_max:
            The upper limit of the exclusive range of random integers.
            Valid value range: 1 to 200,000. Default: 0
        :param timeout:
            The duration in ms the solver
            runs before exiting. If the value is set
            to 0, it does not time out. Default: 0
        """
        super().__init__(
            workspace=workspace,
            provider_id="1qbit",
            name=name,
            input_data_format="microsoft.qio.v2",
            output_data_format="microsoft.qio-results.v1",
            nested_params=False,
            force_str_params=True,
            **kwargs
        )

        self.set_one_param("improvement_cutoff", improvement_cutoff)
        self.set_one_param("improvement_tolerance", improvement_tolerance)
        self.set_one_param("seed", seed)
        self.set_one_param("tabu_tenure", tabu_tenure)
        self.set_one_param("tabu_tenure_rand_max", tabu_tenure_rand_max)
        self.set_one_param("timeout", timeout)


class PticmSolver(Solver):
    target_names = ("1qbit.pticm",)

    def __init__(
        self,
        workspace: Workspace,
        name: str = "1qbit.pticm",
        auto_set_temperatures: Optional[bool] = None,
        elite_threshold: Optional[float] = None,
        frac_icm_thermal_layers: Optional[float] = None,
        frac_sweeps_fixing: Optional[float] = None,
        frac_sweeps_idle: Optional[float] = None,
        frac_sweeps_stagnation: Optional[float] = None,
        goal: Optional[str] = None,
        high_temp: Optional[float] = None,
        low_temp: Optional[float] = None,
        max_samples_per_layer: Optional[int] = None,
        max_total_sweeps: Optional[int] = None,
        manual_temperatures: Optional[List[float]] = None,
        num_elite_temps: Optional[int] = None,
        num_replicas: Optional[int] = None,
        num_sweeps_per_run: Optional[int] = None,
        num_temps: Optional[int] = None,
        perform_icm: Optional[bool] = None,
        scaling_type: Optional[str] = None,
        seed: Optional[int] = None,
        var_fixing_type: Optional[str] = None,
        **kwargs
    ):
        """The constructor of a PTICM solver.

        The parallel tempering with isoenergetic
        cluster moves (PTICM) solver is
        a Monte Carlo approach to solving QUBO problems.
        In this algorithm, multiple
        replicas of the original system,
        each with a different initial state, are
        simulated at different temperatures simultaneously. The replicas at
        neighboring temperatures are periodically swapped based on a Metropolis
        criterion. These swaps allow
        different replicas to do a random walk in the
        temperature space, thereby, efficiently overcoming energy barriers.

        For more information please visit:
        https://portal.1qbit.com/docs/model/pticmsolver

        :param auto_set_temperatures:
            This defines whether the temperature parameters are auto-calculated
            or not. Set it to True for
            auto-calculating and False for customizing
            the temperature parameters. Default: True
        :param elite_threshold:
            The fraction of the best solutions used for the var_fixing_type
            parameter with value SPVAR. Default: 0.3
        :param frac_icm_thermal_layers:
            The fraction of temperatures for the iso-energetic cluster moves.
            Default: 0.5
        :param frac_sweeps_fixing:
            The fraction of sweeps used for fixing the QUBO variables.
            Default: 0.15
        :param frac_sweeps_idle:
            The fraction of sweeps to wait before fixing the QUBO variables.
            Default: 1.0
        :param frac_sweeps_stagnation:
            The fraction of sweeps without improvement that triggers a restart.
            Default: 1.0
        :param goal:
            This defines whether the solver is used for optimizing or sampling.
            Valid values: "OPTIMIZE" or "SAMPLE". Default: "OPTIMIZE"
        :param high_temp:
            The highest temperature of a replica. Set the auto_set_temperatures
            parameter to False to use this feature. Default: 2
        :param low_temp:
            The lowest temperature of a replica. Set the auto_set_temperatures
            parameter to False to use this feature. Default: 0.2
        :param max_samples_per_layer:
            The maximum number of samples collected per replica. Default: 10
        :param max_total_sweeps:
            The total number of sweeps before termination.
            Default: num_sweeps_per_run * 10
        :param manual_temperatures:
            An array of a custom temperature schedule which includes the high,
            intermediate, and low temperature values for the replicas. Set the
            auto_set_temperatures parameter to False to use this feature.
        :param num_elite_temps:
            The number of elite temperatures used for fixing the variables with
            persistency. Default = 4
        :param num_replicas:
            The number of replicas at each temperature. Default: 2
        :param num_sweeps_per_run:
            The number of Monte Carlo sweeps. Default: 100
        :param num_temps:
            The number of temperatures including the highest and the lowest
            temperatures. Set the auto_set_temperatures parameter to False to
            use this feature. Default: 30
        :param perform_icm:
            This parameter is deprecated. Default: None
        :param scaling_type:
            This defines whether the QUBO problem
            is automatically scaled or not.
            "MEDIAN" means it's automatically
            scaled, and "NO_SCALING" means it's
            not. Valid values: "MEDIAN" or "NO_SCALING"
        :param seed:
            The seed in the random number generator to control for deterministic
            results. Default: a random integer
        :param var_fixing_type:
            This decides whether the values of QUBO variables are fixed or not.
            You can fix them with "PERSISTENCY" or
            "SPVAR" types. "NO_FIXING" means the
            variables are not fixed. If you
            choose "PERSISTENCY" or "SPVAR", also
            set the frac_sweeps_fixing and
            frac_sweeps_idle parameters to a number
            less than one. Valid values: "PERSISTENCY", "SPVAR" or "NO_FIXING".
            Default: "NO_FIXING"
        """
        if perform_icm is not None:
            warnings.warn(DeprecationWarning("The perform_icm parameter has been deprecated and will be ignored."))

        super().__init__(
            workspace=workspace,
            provider_id="1qbit",
            name=name,
            input_data_format="microsoft.qio.v2",
            output_data_format="microsoft.qio-results.v1",
            nested_params=False,
            force_str_params=True,
            **kwargs
        )

        self.set_one_param("auto_set_temperatures", auto_set_temperatures)
        self.set_one_param("elite_threshold", elite_threshold)
        self.set_one_param("frac_icm_thermal_layers", frac_icm_thermal_layers)
        self.set_one_param("frac_sweeps_fixing", frac_sweeps_fixing)
        self.set_one_param("frac_sweeps_idle", frac_sweeps_idle)
        self.set_one_param("frac_sweeps_stagnation", frac_sweeps_stagnation)
        self.set_one_param("goal", goal)
        self.set_one_param("high_temp", high_temp)
        self.set_one_param("low_temp", low_temp)
        self.set_one_param("max_samples_per_layer", max_samples_per_layer)
        self.set_one_param("max_total_sweeps", max_total_sweeps)
        self.set_one_param("manual_temperatures", manual_temperatures)
        self.set_one_param("num_elite_temps", num_elite_temps)
        self.set_one_param("num_replicas", num_replicas)
        self.set_one_param("num_sweeps_per_run", num_sweeps_per_run)
        self.set_one_param("num_temps", num_temps)
        self.set_one_param("scaling_type", scaling_type)
        self.set_one_param("seed", seed)
        self.set_one_param("var_fixing_type", var_fixing_type)


class PathRelinkingSolver(Solver):
    target_names = ("1qbit.pathrelinking",)

    def __init__(
        self,
        workspace: Workspace,
        name: str = "1qbit.pathrelinking",
        distance_scale: Optional[float] = None,
        greedy_path_relinking: Optional[bool] = None,
        ref_set_count: Optional[int] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        **kwargs
    ):
        """The constructor of a Path-relinking Search solver.

        The path-relinking algorithm is a heuristic algorithm that uses the
        tabu search as a subroutine to solve a QUBO problem. The algorithm
        starts from a set of elite solutions found by the tabu search. It
        then constructs a path between each pair of elite solutions, selects
        one of the solutions along the path, and repeats the tabu search. If
        the tabu search finds a distinct solution that is better than the
        current worst elite solution, the elite solutions set is updated with
        the new improved solution. This whole procedure is repeated until the
        algorithm meets a stopping condition.

        For more information please visit:
        https://portal.1qbit.com/docs/model/pathrelinkingsolver

        :param distance_scale:
            The minimum distance from the initiating and guiding solutions for
            constructing the candidate solution list. The highest quality
            solution in the candidate solution list is then selected for
            improvement. Valid values: 0.0 to 0.5. Default: 0.33
        :param greedy_path_relinking:
            When you use the path-relinking solver there are two ways you can
            generate a path that leads to the solution:
            one is the greedy function
            and the other operates in a random matter. If you set the this
            parameter to true, the solver will use the greedy function. If you
            set it to false, it will use the random method. Default: False
        :param ref_set_count:
            The number of initial elite solutions to be generated by the tabu
            search algorithm. Valid values: Greater than 1. Default: 10
        :param seed:
            The seed in the random number generator to control for deterministic
            results. Default: a random integer
        :param timeout:
            The duration in ms the solver runs
            before exiting. If the value is set
            to 0, it does not time out. Default: 0
        """
        super().__init__(
            workspace=workspace,
            provider_id="1qbit",
            name=name,
            input_data_format="microsoft.qio.v2",
            output_data_format="microsoft.qio-results.v1",
            nested_params=False,
            force_str_params=True,
            **kwargs
        )

        self.set_one_param("distance_scale", distance_scale)
        self.set_one_param("greedy_path_relinking", greedy_path_relinking)
        self.set_one_param("ref_set_count", ref_set_count)
        self.set_one_param("seed", seed)
        self.set_one_param("timeout", timeout)
